import logging
from collections.abc import Iterable
from typing import TYPE_CHECKING, Any, Optional

if TYPE_CHECKING:
    from google._upb._message import Descriptor, Message

import grpc
from google.protobuf.descriptor_pool import DescriptorPool
from google.protobuf.message_factory import GetMessageClass
from grpc_reflection.v1alpha.proto_reflection_descriptor_database import (
    ProtoReflectionDescriptorDatabase,
)

log = logging.getLogger(__name__)


class GrpcClient:
    """GRPC client for a caikit nlp runtime server

    Args:
        channel (grpc.Channel): a connected GRPC channel for use of making the calls.
    """

    def __init__(
        self,
        host: str,
        port: int,
        *,
        insecure: bool = False,
        ca_cert: Optional[bytes] = None,
        client_key: Optional[bytes] = None,
        client_cert: Optional[bytes] = None,
    ) -> None:
        """Client class for a Caikit NLP grpc server

        >>> # To connect via an insecure port
        >>> client = GrpcClient("localhost", port=8085)
        >>> generated_text = client.generate_text_stream(
        >>>     "flan-t5-small-caikit",
        >>>     "What is the boiling point of Nitrogen?",
        >>> )
        """

        try:
            self._channel = self._make_channel(
                host,
                port,
                insecure=insecure,
                client_key=client_key,
                client_cert=client_cert,
                ca_cert=ca_cert,
            )
        except grpc._channel._MultiThreadedRendezvous as exc:
            log.error("Could not connect to the server: %s", exc.details)
            raise RuntimeError(f"Could not connect to {host}:{port}") from None

        self._reflection_db = ProtoReflectionDescriptorDatabase(self._channel)
        self._desc_pool = DescriptorPool(self._reflection_db)
        try:
            self._text_generation_task_request = GetMessageClass(
                self._desc_pool.FindMessageTypeByName(
                    "caikit.runtime.Nlp.TextGenerationTaskRequest"
                )
            )
            self._task_text_generation_request = GetMessageClass(
                self._desc_pool.FindMessageTypeByName(
                    "caikit.runtime.Nlp.ServerStreamingTextGenerationTaskRequest"
                )
            )
            self._generated_text_result = GetMessageClass(
                self._desc_pool.FindMessageTypeByName(
                    "caikit_data_model.nlp.GeneratedTextResult"
                )
            )
            self._task_predict = self._channel.unary_unary(
                "/caikit.runtime.Nlp.NlpService/TextGenerationTaskPredict",
                request_serializer=self._text_generation_task_request.SerializeToString,
                response_deserializer=self._generated_text_result.FromString,
            )
            self._streaming_task_predict = self._channel.unary_stream(
                "/caikit.runtime.Nlp.NlpService/ServerStreamingTextGenerationTaskPredict",
                request_serializer=self._task_text_generation_request.SerializeToString,
                response_deserializer=self._generated_text_result.FromString,
            )
        except KeyError as exc:
            log.error("The grpc server does not have the type: %s", exc)
            raise ValueError(str(exc)) from exc

    def generate_text(self, model_id: str, text: str, **kwargs) -> str:
        """Sends a generate text request to the server for the given model id

        Args:
            model_id: the model identifier
            text: the text to generate

        Keyword Args:
            preserve_input_text (Bool): preserve the input text (default to False)
            max_new_tokens (int): maximum number of new tokens
            min_new_tokens (int): minimum number of new tokens

        Raises:
            ValueError: thrown if an empty model id is passed
            exc: thrown if any exceptions are caught while creating and sending
            the text generation request

        Returns:
            the generated text
        """
        if model_id == "":
            raise ValueError("request must have a model id")

        log.info(f"Calling generate_text for '{model_id}'")
        metadata = [("mm-model-id", model_id)]

        request = self._text_generation_task_request()

        self._populate_request(request, text, **kwargs)
        try:
            response = self._task_predict(request=request, metadata=metadata)
        except grpc._channel._InactiveRpcError as exc:
            raise RuntimeError(exc.details()) from None

        log.debug(f"Response: {response}")
        result = response.generated_text
        log.info("Calling generate_text was successful")
        return result

    def get_text_generation_parameters(self) -> dict[str, Any]:
        """returns a dict with available fields and their type"""
        descriptor: Descriptor = self._task_text_generation_request.DESCRIPTOR

        # hack: used map the grpc type (int) to a human-readable string
        grpc_type_to_str = {
            getattr(descriptor.fields[0], t): t.split("_")[1].lower()
            for t in dir(descriptor.fields[0])
            if t.startswith("TYPE_")
        }

        def simplify_descriptor(
            descriptor: "Descriptor",
        ) -> dict:
            """recursively flattens grpc descriptor into a human-friendly dict"""

            flattened: dict = {}
            for field in descriptor.fields:
                string_type = grpc_type_to_str[field.type]

                if field.message_type:
                    flattened[field.name] = simplify_descriptor(field.message_type)
                else:
                    flattened[field.name] = string_type

            return flattened

        return simplify_descriptor(descriptor)

    def __enter__(self):
        return self

    def __exit__(self, *exc_info):
        self._close()
        return False

    def generate_text_stream(self, model_id: str, text: str, **kwargs) -> Iterable[str]:
        """Sends a generate text stream request to the server for the given model id

        Args:
            model_id: the model identifier
            text: the text to generate

        Keyword Args:
            preserve_input_text (Bool): preserve the input text (default to False)
            max_new_tokens (int): maximum number of new tokens
            min_new_tokens (int): minimum number of new tokens

        Raises:
            ValueError: thrown if an empty model id is passed

        Returns:
            a list of generated text (tokens)

        Example:

        >>> text = "What is 2+2?"
        >>> chunks = []
        >>> for chunk in grpc_client.generate_text_stream(
        >>>     "flan-t5-small-caikit",
        >>>     text,
        >>> ):
        >>>     print(f"Got {chunk=}")
        >>>     chunks.append(chunk)
        >>> print(f"final result: {''.join(chunks)}")

        """
        if not model_id:
            raise ValueError("request must have a model id")

        log.info(f"Calling generate_text_stream for '{model_id}'")

        metadata = [("mm-model-id", model_id)]

        request = self._task_text_generation_request()
        self._populate_request(request, text, **kwargs)

        try:
            yield from (
                message.generated_text
                for message in self._streaming_task_predict(
                    metadata=metadata, request=request
                )
            )
        except grpc._channel._MultiThreadedRendezvous as exc:
            raise RuntimeError(exc.details()) from None

    def _populate_request(self, request: "Message", text: str, **kwargs):
        """dynamically converts kwargs to request attributes."""
        request.text = text
        for key, value in kwargs.items():
            try:
                setattr(request, key, value)
            except AttributeError as exc:
                raise ValueError(f"Unsupported kwarg {key=}") from exc

    def _close(self):
        try:
            if hasattr(self, "_channel") and self._channel:
                self._channel.close()
        except Exception:
            log.exception("Unexpected exception while closing client")

    def __del__(self):
        self._close()

    def _make_channel(
        self,
        host: str,
        port: int,
        *,
        insecure: bool = False,
        ca_cert: Optional[bytes] = None,
        client_key: Optional[bytes] = None,
        client_cert: Optional[bytes] = None,
    ) -> grpc.Channel:
        if not host.strip():
            raise ValueError("A non empty host name is required")
        if int(port) <= 0:
            raise ValueError("A non zero port number is required")

        connection = f"{host}:{port}"
        if insecure:
            log.warning("Connecting over an insecure grpc channel")
            return grpc.insecure_channel(connection)

        credentials_kwargs: dict[str, bytes] = {}
        if ca_cert and not (any((client_cert, client_key))):
            log.info("Connecting using provided CA certificate for secure channel")
            credentials_kwargs.update(root_certificates=ca_cert)
        elif client_cert and client_key and ca_cert:
            log.info("Connecting using mTLS for secure channel")
            credentials_kwargs.update(
                root_certificates=ca_cert,
                private_key=client_key,
                certificate_chain=client_cert,
            )
        else:
            raise ValueError("mTLS requires client_cert, client_key and ca_cert")

        return grpc.secure_channel(
            connection,
            grpc.ssl_channel_credentials(**credentials_kwargs),
        )



import pytest
import requests
from caikit_nlp_client.http_client import HttpClient

from .utils import ConnectionType, get_random_port, wait_until


@pytest.fixture(scope="session")
def http_server_thread_port():
    """port for caikit http runtime thread"""
    return get_random_port()


@pytest.fixture(scope="session")
def using_real_caikit(pytestconfig):
    return pytestconfig.option.real_caikit


@pytest.fixture(scope="session")
def http_client(
    caikit_nlp_runtime,
    http_server,
    using_real_caikit,
    request: pytest.FixtureRequest,
    connection_type,
    client_cert_file,
    client_key_file,
    ca_cert_file,
) -> HttpClient:
    if using_real_caikit:
        if connection_type is not ConnectionType.INSECURE:
            pytest.skip(reason="not testing TLS with a docker caikit instance")

        host, port = request.getfixturevalue("http_server_docker")
    else:
        host, port = request.getfixturevalue("http_server_thread")

    kwargs: dict[str, str] = {}

    if connection_type is ConnectionType.INSECURE:
        url = f"http://{host}:{port}"
    elif connection_type is ConnectionType.TLS:
        # a valid certificate autority should validate the response with no extra args
        url = f"https://{host}:{port}"

    elif ConnectionType.MTLS:
        url = f"https://{host}:{port}"
        kwargs.update(
            ca_cert_path=ca_cert_file,
            client_cert_path=client_cert_file,
            client_key_path=client_key_file,
        )
    else:
        raise ValueError(f"invalid {connection_type=}")

    return HttpClient(url, **kwargs)


@pytest.fixture(scope="function")
def accept_self_signed_certs(ca_cert_file, monkeysession):
    """validates self signed certs by injecting REQUEST_CA_BUNDLE"""
    with monkeysession.context() as monkeypatch:
        monkeypatch.setenv("REQUESTS_CA_BUNDLE", ca_cert_file)

        yield


@pytest.fixture(scope="session")
def http_server_thread(
    caikit_nlp_runtime,
    http_server_thread_port,
    mock_text_generation,
):
    """spins a caikit http server in a thread for testing, returning host and port"""
    from caikit.runtime.http_server import RuntimeHTTPServer

    http_server = RuntimeHTTPServer()
    http_server.start(blocking=False)

    yield "localhost", http_server_thread_port

    http_server.stop()


@pytest.fixture(scope="session")
def http_server(
    pytestconfig,
    request: pytest.FixtureRequest,
    caikit_nlp_runtime,
    mock_text_generation,
    connection_type,
    ca_cert_file,
    client_cert_file,
    client_key_file,
):
    if pytestconfig.option.real_caikit:
        if connection_type is not ConnectionType.INSECURE:
            pytest.skip(reason="not testing TLS with a docker caikit instance")

        host, port = request.getfixturevalue("http_server_docker")
    else:
        host, port = request.getfixturevalue("http_server_thread")

    def health_check(host: str, port: int):
        if connection_type is ConnectionType.INSECURE:
            scheme = "http"
            kwargs = {}
        elif connection_type is ConnectionType.TLS:
            scheme = "https"
            kwargs = {"verify": ca_cert_file}
        elif connection_type is ConnectionType.MTLS:
            scheme = "https"
            kwargs = {
                "verify": ca_cert_file,
                "cert": (client_cert_file, client_key_file),
            }
        else:
            raise ValueError(f"Invalid {connection_type=}")
        response = requests.get(f"{scheme}://{host}:{port}/health", **kwargs)
        assert response.status_code == 200
        assert response.text == "OK"

    wait_until(
        lambda: health_check(host, port),
        pause=0.5,
    )

    yield host, port



from collections.abc import Iterable

import pytest


@pytest.fixture(scope="session")
def caikit_test_producer():
    from caikit.interfaces.nlp.data_model.text_generation import ProducerId

    yield ProducerId(
        name="Testing Producer",
        version="0.0.1",
    )


@pytest.fixture(scope="session")
def prompt():
    yield "What does foobar mean?"


@pytest.fixture(scope="session")
def generated_text():
    yield "a symphony"


@pytest.fixture(scope="session")
def generated_text_result(caikit_test_producer, generated_text):
    from caikit.interfaces.nlp.data_model.text_generation import (
        FinishReason,
        GeneratedTextResult,
    )

    yield GeneratedTextResult(
        generated_text=generated_text,
        generated_tokens=42,
        finish_reason=FinishReason.EOS_TOKEN,
        producer_id=caikit_test_producer,
        input_token_count=10,
        seed=None,  # Optional[np.uint64]
    )


# FIXME: Validate text stream mocking. There's a lot of logic here.
#        Can this be simplified?
@pytest.fixture(scope="session")
def generated_text_stream_result(caikit_test_producer, generated_text, prompt):
    from caikit.interfaces.nlp.data_model.text_generation import (
        FinishReason,
        GeneratedTextStreamResult,
        GeneratedToken,
        TokenStreamDetails,
    )

    token_list = [GeneratedToken(text=prompt, logprob=0.42)]
    input_token_count = len(token_list)

    generated_tokens = ["", "", "a", " ", "s", "y", "m", "phon", "y", ""]
    assert (
        "".join(generated_tokens) == generated_text
    ), "generated_tokens should match the generated_text"

    result = []
    for token in generated_tokens:
        details = TokenStreamDetails(
            finish_reason=FinishReason.NOT_FINISHED,
            generated_tokens=42,  # FIXME: is this correct?
            seed=None,
            input_token_count=input_token_count,
        )

        stream_result = GeneratedTextStreamResult(
            generated_text=token,
            tokens=token_list,
            details=details,
        )

        result.append(stream_result)

    result[-1].details.finish_reason = FinishReason.EOS_TOKEN
    return result


@pytest.fixture(scope="session")
def mock_text_generation(
    request: pytest.FixtureRequest,
    generated_text_result,
    generated_text_stream_result,
    monkeysession,
):
    # import caikit_nlp.modules.text_generation.text_generation_local
    import caikit_nlp.modules.text_generation.text_generation_tgis
    from caikit.interfaces.nlp.data_model.text_generation import (
        GeneratedTextResult,
        GeneratedTextStreamResult,
    )

    # NOTE: config uses tgis, so this is not really required,
    #       unless we want to test the local text generation module
    # monkeypatch.setattr(
    #     caikit_nlp.modules.text_generation.text_generation_local,
    #     "generate_text_func",
    #     lambda: generated_text_result,
    # )

    class StubTGISGenerationClient:
        """stub TGISGenerationClient

        `raise_exception` can be passed as kwarg with any value to raise
        an exception for a text generation  request

        """

        def __init__(self, *args, **kwargs):
            pass

        def unary_generate(self, *args, **kwargs) -> GeneratedTextResult:
            if "[[raise exception]]" in kwargs["text"]:
                raise ValueError("user requested an exception")

            return generated_text_result

        def stream_generate(
            self, *args, **kwargs
        ) -> Iterable[GeneratedTextStreamResult]:
            if "[[raise exception]]" in kwargs["text"]:
                raise ValueError("user requested an exception")

            yield from generated_text_stream_result

    monkeysession.setattr(
        caikit_nlp.modules.text_generation.text_generation_tgis,
        "TGISGenerationClient",
        StubTGISGenerationClient,
    )

    yield



from pathlib import Path

import pytest


@pytest.fixture(scope="session")
def ca_cert_file():
    return str(Path(__file__).parent / "resources/ssl-certs/ca.pem")


@pytest.fixture(scope="session")
def ca_cert(ca_cert_file):
    with open(ca_cert_file, "rb") as fh:
        return fh.read()


@pytest.fixture(scope="session")
def client_key_file():
    return str(Path(__file__).parent / "resources/ssl-certs/client-key.pem")


@pytest.fixture(scope="session")
def client_key(client_key_file):
    with open(client_key_file, "rb") as fh:
        return fh.read()


@pytest.fixture(scope="session")
def client_cert_file():
    return str(Path(__file__).parent / "resources/ssl-certs/client.pem")


@pytest.fixture(scope="session")
def client_cert(client_cert_file):
    with open(client_cert_file, "rb") as fh:
        return fh.read()


@pytest.fixture(scope="session")
def server_key_file():
    return str(Path(__file__).parent / "resources/ssl-certs/server-key.pem")


@pytest.fixture(scope="session")
def server_key(server_key_file):
    with open(server_key_file, "rb") as fh:
        return fh.read()


@pytest.fixture(scope="session")
def server_cert_file():
    return str(Path(__file__).parent / "resources/ssl-certs/server.pem")


@pytest.fixture(scope="session")
def server_cert(server_cert_file):
    with open(server_cert_file, "rb") as fh:
        return fh.read()



import json
import logging
from collections.abc import Iterable
from typing import Any, Optional, Union

import requests

log = logging.getLogger(__name__)


class HttpClient:
    """HTTP client for a caikit nlp runtime server

    Args:
        http_config (HttpConfig): Configurations to make HTTP call.
    """

    def __init__(
        self,
        base_url: str,
        ca_cert_path: Optional[str] = None,
        client_cert_path: Optional[str] = None,
        client_key_path: Optional[str] = None,
    ):
        """Client class for a Caikit NLP HTTP server

        For unsecured connections:

        >>> client = HttpClient("http://localhost:8080")

        For TLS (https) connections:

        >>> client = HttpClient("https://localhost:8080")

        For mTLS connections, both client_cert_path and client_key_path have to be
        provided:

        >>> client = HttpClient("https://localhost:8080",
        >>>     ca_cert_path="path/to/ca.pem",
        >>>     client_cert_path='path/to/client_cert.pem',
        >>>     client_key_path='path/to/client_key.pem'
        >>> )

        """
        text_generation_endpoint = "/api/v1/task/text-generation"
        text_generation_stream_endpoint = (
            "/api/v1/task/server-streaming-text-generation"
        )

        self._api_base = base_url
        self._api_url = f"{base_url}{text_generation_endpoint}"
        self._stream_api_url = f"{base_url}{text_generation_stream_endpoint}"

        self._client_cert_path = client_cert_path
        self._client_key_path = client_key_path
        self._ca_cert_path = ca_cert_path
        if (
            any(
                (
                    self._client_key_path,
                    self._client_cert_path,
                    self._ca_cert_path,
                )
            )
            and not self._mtls_configured
        ):
            raise ValueError(
                "Must provide both ca_cert_path, client_cert_path, client_key_path "
                "for mTLS"
            )

    @property
    def _mtls_configured(self):
        return all((self._client_key_path, self._client_cert_path, self._ca_cert_path))

    def _get_tls_configuration(self) -> dict[str, Union[str, tuple[str, str]]]:
        req_kwargs: dict[str, Union[str, tuple[str, str]]] = {}
        if self._mtls_configured:
            assert self._client_key_path
            assert self._client_cert_path

            req_kwargs["cert"] = (
                self._client_cert_path,
                self._client_key_path,
            )
        if self._ca_cert_path:
            req_kwargs["verify"] = self._ca_cert_path

        return req_kwargs

    def get_text_generation_parameters(self, timeout: float = 60.0) -> dict[str, str]:
        """returns a dict with available fields and their type"""
        req_kwargs = self._get_tls_configuration()

        openapi_spec = requests.get(
            f"{self._api_base}/openapi.json",
            timeout=timeout,
            **req_kwargs,  # type: ignore
        ).json()

        request_schema = openapi_spec["paths"]["/api/v1/task/text-generation"]["post"][
            "requestBody"
        ]["content"]["application/json"]["schema"]
        parameters = request_schema["properties"]["parameters"]

        def simplify_parameter_schema(parameters: dict) -> dict:
            """recursively flattens openapi's spec into a human-friendly dict"""
            assert not len(parameters) > 1, "parameters should be a list of 1 dict 🤷"

            flattened = {}
            for param, description in parameters["allOf"][0]["properties"].items():
                if "allOf" in description:
                    flattened[param] = simplify_parameter_schema(description)
                else:
                    flattened[param] = description["type"]

            return flattened

        return simplify_parameter_schema(parameters)

    def generate_text(
        self, model_id: str, text: str, timeout: float = 60.0, **kwargs
    ) -> str:
        """Queries the `text-generation` endpoint for the given model_id

        Args:
            model_id: the model identifier
            text: the text to generate

        Keyword Args:
            preserve_input_text (Bool): preserve the input text (default to False)
            max_new_tokens (int): maximum number of new tokens
            min_new_tokens (int): minimum number of new tokens
            timeout (int): HTTP request timeout value in seconds

        Raises:
            ValueError: thrown if an empty model id is passed
            exc: thrown if any exceptions are caught while creating and sending
            the text generation request

        Returns:
            the generated text

        Example:

        >>> generated_text = client.generate_text(
                "flan-t5-small-caikit",
                "What is the boiling point of Nitrogen?"
            )
        """
        if not model_id:
            raise ValueError("request must have a model id")

        log.info(f"Calling generate_text for '{model_id}'")
        json_input = self._create_json_request(
            model_id,
            text,
            **kwargs,
        )

        req_kwargs = self._get_tls_configuration()

        response = requests.post(
            self._api_url,
            json=json_input,
            timeout=timeout,
            **req_kwargs,  # type: ignore
        )
        log.debug(f"Response: {response}")
        if response.status_code == 200:
            return response.json()["generated_text"]
        elif 400 <= response.status_code < 500:
            raise RuntimeError(response.json()["details"])
        else:
            raise RuntimeError(
                f"{response.status_code}: Server error {response.reason}"
            )

    def generate_text_stream(
        self, model_id: str, text: str, timeout: float = 60.0, **kwargs
    ) -> Iterable[str]:
        """Queries the `text-generation` stream endpoint for the given model_id

        Args:
            model_id: the model identifier
            text: the text to generate

        Keyword Args:
            preserve_input_text (Bool): preserve the input text (default to False)
            max_new_tokens (int): maximum number of new tokens
            min_new_tokens (int): minimum number of new tokens
            timeout (int): HTTP request timeout value in seconds

        Raises:
            ValueError: thrown if an empty model id is passed
            exc: thrown if any exceptions are caught while creating and sending
            the text generation request

        Returns:
            a list of messages generated from the server

        Example:

        >>> text = "What is 2+2?"
        >>> chunks = []
        >>> for chunk in http_client.generate_text_stream(
        >>>     "flan-t5-small-caikit",
        >>>     text,
        >>> ):
        >>>     print(f"Got {chunk=}")
        >>>     chunks.append(chunk)
        >>> print(f"final result: {''.join(chunks)}")

        """
        if not model_id:
            raise ValueError("request must have a model id")

        log.info(f"Calling generate_text_stream for '{model_id}'")

        payload: dict[str, Any] = {
            "model_id": model_id,
            "inputs": text,
        }
        if kwargs:
            payload["parameters"] = kwargs

        req_kwargs = self._get_tls_configuration()

        response = requests.post(
            self._stream_api_url,
            json=payload,
            timeout=timeout,
            **req_kwargs,  # type: ignore
        )
        log.debug(f"Response: {response}")
        buffer: list[bytes] = []
        for line in response.iter_lines():
            if line:
                # the first 6 bytes contain "data: ", we can skip those
                buffer.append(line[6:])
                continue

            try:
                message = json.loads(b"".join(buffer))
            except json.JSONDecodeError:
                # message not over yet
                continue

            buffer.clear()
            if "details" in message and "code" in message:
                raise RuntimeError(
                    "Exception iterating responses: {}".format(message["details"])
                )
            yield message["generated_text"]

        if buffer:
            final_message = json.loads(b"".join(buffer))
            if "details" in final_message and "code" in final_message:
                raise RuntimeError(
                    "Exception iterating responses: {}".format(final_message["details"])
                )
            try:
                yield final_message["generated_text"]
            except KeyError as exc:
                raise RuntimeError(
                    "Unexpected response from the server: generated text is missing"
                ) from exc

    def _create_json_request(self, model_id, text, **kwargs) -> dict[str, Any]:
        json_input = {
            "model_id": model_id,
            "inputs": text,
            "parameters": {"max_new_tokens": 200, "min_new_tokens": 10},
        }
        if parameters := json_input.get("parameters"):
            parameters.update(kwargs)
        return json_input



from caikit_nlp_client import GrpcClient, HttpClient

host = "localhost"
http_port = 8080
grpc_port = 8085
text = "What is the boiling point of Nitrogen?"

# Using the http client
http_client = HttpClient(f"http://{host}:{http_port}")

generated_text = http_client.generate_text("flan-t5-small-caikit", text)
print(generated_text)
# {"generated_text": "...", "generated_tokens": 15, "finish_reason": "EOS_TOKEN", "producer_id": {"name": "Text Generation", "version": "0.1.0"}, "input_token_count": 11, "seed": null}

http_client.generate_text(
    "flan-t5-small-caikit", text, min_new_tokens=20, max_new_tokens=20
)
print(generated_text)
# {"generated_text": "...", "generated_tokens": 20, "finish_reason": "MAX_TOKENS", "producer_id": {"name": "Text Generation", "version": "0.1.0"}, "input_token_count": 11, "seed": null}


# using the grpc client
grpc_client = GrpcClient(host=host, port=grpc_port, insecure=True)
text = grpc_client.generate_text(
    "flan-t5-small-caikit", text, min_new_tokens=20, max_new_tokens=20
)
print(text)
# 'this is the generated text'

# using a context manager (grpc only):
# (makes sure that the underlying resources are cleaned)
with GrpcClient(host, port=grpc_port, insecure=True) as client:
    text = client.generate_text(
        "flan-t5-small-caikit", text, min_new_tokens=20, max_new_tokens=20
    )


# using the streaming implementation, grpc
chunks = []
for chunk in grpc_client.generate_text_stream("flan-t5-small-caikit", text):
    print(f"Got {chunk=}")
    chunks.append(chunk)
print(f"generated text: {''.join(chunks)}")

# using the streaming implementation, http
chunks = []
for chunk in http_client.generate_text_stream("flan-t5-small-caikit", text):
    print(f"Got {chunk=}")
    chunks.append(chunk)
print(f"generated text: {''.join(chunks)}")


# Using a self-signed CA Certificate
http_client = HttpClient(f"https://{host}:{http_port}", ca_cert_path="ca.pem")

with open("ca.pem", "rb") as fh:
    ca_cert = fh.read()
grpc_client = GrpcClient(host, grpc_port, ca_cert=ca_cert)



from .grpc_client import GrpcClient
from .http_client import HttpClient

try:
    from ._version import __version__
except ImportError:  # pragma: no cover
    __version__ = "unknown"

__all__ = ["GrpcClient", "HttpClient"]



"""Nox sessions."""
import glob
import os
import shlex
import sys
from pathlib import Path
from textwrap import dedent

import nox

package = "caikit_nlp_client"
python_versions = ["3.11", "3.10", "3.9"]
nox.needs_version = ">= 2021.6.6"
nox.options.sessions = ("pre-commit", "mypy", "tests", "build")


# see note regarding caikit_nlp in pyproject.toml
caikit_nlp_version = "caikit-nlp @ git+https://github.com/caikit/caikit-nlp@0.3.0"


def activate_virtualenv_in_precommit_hooks(session: nox.Session) -> None:
    """Activate virtualenv in hooks installed by pre-commit.

    This function patches git hooks installed by pre-commit to activate the
    session's virtual environment. This allows pre-commit to locate hooks in
    that environment when invoked from git.

    Args:
        session: The Session object.
    """
    assert session.bin is not None  # nosec

    # Only patch hooks containing a reference to this session's bindir. Support
    # quoting rules for Python and bash, but strip the outermost quotes so we
    # can detect paths within the bindir, like <bindir>/python.
    bindirs = [
        bindir[1:-1] if bindir[0] in "'\"" else bindir
        for bindir in (repr(session.bin), shlex.quote(session.bin))
    ]

    virtualenv = session.env.get("VIRTUAL_ENV")
    if virtualenv is None:
        return

    headers = {
        # pre-commit < 2.16.0
        "python": f"""\
            import os
            os.environ["VIRTUAL_ENV"] = {virtualenv!r}
            os.environ["PATH"] = os.pathsep.join((
                {session.bin!r},
                os.environ.get("PATH", ""),
            ))
            """,
        # pre-commit >= 2.16.0
        "bash": f"""\
            VIRTUAL_ENV={shlex.quote(virtualenv)}
            PATH={shlex.quote(session.bin)}"{os.pathsep}$PATH"
            """,
        # pre-commit >= 2.17.0 on Windows forces sh shebang
        "/bin/sh": f"""\
            VIRTUAL_ENV={shlex.quote(virtualenv)}
            PATH={shlex.quote(session.bin)}"{os.pathsep}$PATH"
            """,
    }

    hookdir = Path(".git") / "hooks"
    if not hookdir.is_dir():
        return

    for hook in hookdir.iterdir():
        if hook.name.endswith(".sample") or not hook.is_file():
            continue

        if not hook.read_bytes().startswith(b"#!"):
            continue

        text = hook.read_text()

        if not any(
            Path("A") == Path("a") and bindir.lower() in text.lower() or bindir in text
            for bindir in bindirs
        ):
            continue

        lines = text.splitlines()

        for executable, header in headers.items():
            if executable in lines[0].lower():
                lines.insert(1, dedent(header))
                hook.write_text("\n".join(lines))
                break


@nox.session(name="pre-commit")
def precommit(session: nox.Session) -> None:
    """Lint using pre-commit."""
    args = session.posargs or [
        "run",
        "--all-files",
        "--hook-stage=manual",
        "--show-diff-on-failure",
    ]
    session.install(
        "pre-commit",
        "pre-commit-hooks",
        "ruff",
        "pyupgrade",
    )
    session.run("pre-commit", *args)
    if args and args[0] == "install":
        activate_virtualenv_in_precommit_hooks(session)


@nox.session
def mypy(session: nox.Session) -> None:
    """Type-check using mypy."""
    args = session.posargs or ["src", "tests"]
    session.install(
        "--index-url=https://download.pytorch.org/whl/cpu", "torch"
    )  # use torch-cpu to speed up tests
    session.install(caikit_nlp_version)
    session.install(".[dev,tests]")
    session.run("python", "-m", "mypy", *args)
    if not session.posargs:
        session.run("mypy", f"--python-executable={sys.executable}", "noxfile.py")


@nox.session(python=python_versions)
def tests(session: nox.Session) -> None:
    """Run the test suite."""
    session.install(
        "--index-url=https://download.pytorch.org/whl/cpu", "torch"
    )  # use torch-cpu to speed up tests
    session.install(caikit_nlp_version)
    session.install(".[tests]")
    try:
        session.run(
            "pytest",
            "--cov",
            "--cov-config=pyproject.toml",
            *session.posargs,
            env={"COVERAGE_FILE": f".coverage.{session.python}"},
        )
    finally:
        if session.interactive:
            session.notify("coverage", posargs=[])


@nox.session
def coverage(session: nox.Session) -> None:
    """Produce the coverage report."""
    args = session.posargs or ["report"]

    session.install("coverage[toml]")

    if not session.posargs and any(Path().glob(".coverage.*")):
        session.run("coverage", "combine")

    session.run("coverage", *args)


@nox.session
def build(session: nox.Session) -> None:
    session.install("build", "setuptools", "twine")
    session.run("python", "-m", "build")
    dists = glob.glob("dist/*")
    session.run("twine", "check", *dists, silent=True)



from typing import Optional

import grpc
import pytest
from caikit_nlp_client.grpc_client import GrpcClient
from grpc_health.v1 import health_pb2, health_pb2_grpc

from .utils import ConnectionType, get_random_port, wait_until


def channel_factory(
    host: str,
    port: int,
    connection_type: ConnectionType,
    ca_cert: Optional[bytes] = None,
    client_key: Optional[bytes] = None,
    client_cert: Optional[bytes] = None,
) -> grpc.Channel:
    connection = f"{host}:{port}"
    if connection_type is ConnectionType.INSECURE:
        return grpc.insecure_channel(connection)
    if connection_type is ConnectionType.TLS:
        return grpc.secure_channel(
            connection,
            grpc.ssl_channel_credentials(ca_cert),
        )
    if connection_type is ConnectionType.MTLS:
        return grpc.secure_channel(
            connection,
            grpc.ssl_channel_credentials(ca_cert, client_key, client_cert),
        )


@pytest.fixture(scope="session")
def grpc_server_thread_port():
    """port for caikit grpc runtime thread"""
    return get_random_port()


@pytest.fixture(scope="session")
def grpc_client(
    grpc_server,
    connection_type,
    ca_cert,
    client_key,
    client_cert,
) -> GrpcClient:
    if connection_type is ConnectionType.INSECURE:
        return GrpcClient(*grpc_server, insecure=True)

    if connection_type is ConnectionType.TLS:
        return GrpcClient(*grpc_server, ca_cert=ca_cert)

    if connection_type is ConnectionType.MTLS:
        return GrpcClient(
            *grpc_server,
            ca_cert=ca_cert,
            client_key=client_key,
            client_cert=client_cert,
        )

    raise ValueError(f"invalid {connection_type=}")


@pytest.fixture(scope="session")
def grpc_server_thread(
    caikit_nlp_runtime,
    grpc_server_thread_port,
    mock_text_generation,
):
    """spins a caikit grpc server in a thread for testing, returning host and port"""
    from caikit.runtime.grpc_server import RuntimeGRPCServer

    grpc_server = RuntimeGRPCServer()
    grpc_server.start(blocking=False)

    yield "localhost", grpc_server_thread_port

    grpc_server.stop()


@pytest.fixture(scope="session")
def grpc_server(
    pytestconfig,
    request: pytest.FixtureRequest,
    connection_type,
    ca_cert,
    client_key,
    client_cert,
):
    if pytestconfig.option.real_caikit:
        if connection_type is not ConnectionType.INSECURE:
            pytest.skip(reason="not testing TLS with a docker caikit instance")
        host, port = request.getfixturevalue("grpc_server_docker")
    else:
        host, port = request.getfixturevalue("grpc_server_thread")

    def health_check(host: str, port: int):
        if connection_type is ConnectionType.INSECURE:
            kwargs = {}
        elif connection_type is ConnectionType.TLS:
            kwargs = {
                "ca_cert": ca_cert,
            }
        elif connection_type is ConnectionType.MTLS:
            kwargs = {
                "client_cert": client_cert,
                "client_key": client_key,
                "ca_cert": ca_cert,
            }
        else:
            raise ValueError(f"Unknown {connection_type=}")

        channel = channel_factory(host, port, connection_type, **kwargs)
        stub = health_pb2_grpc.HealthStub(channel)
        health_check_request = health_pb2.HealthCheckRequest()
        stub.Check(health_check_request)

    wait_until(
        lambda: health_check(host, port),
        pause=0.5,
    )

    yield host, port
